/*
 * vip1710.c: A plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id: vip1710.c,v 1.9 2007/01/02 20:00:00 dom Exp $
 */

#include <vdr/plugin.h>
#include "vip1710device.h"
#include "vip1710menu.h"
#include "vip1710.h"
#if VDRVERSNUM < 10507
#include "po/i18n.h"
#endif //VDRVERSNUM < 10507

static const char *VERSION        = "0.1.0";
#if VDRVERSNUM < 10507
static const char *DESCRIPTION    = "VIP1710 as output device";
#else
static const char *DESCRIPTION    = trNOOP("VIP1710 as output device");
#endif


cPluginVip1710::cPluginVip1710(void)
{
  // Initialize any member variables here.
  // DON'T DO ANYTHING ELSE THAT MAY HAVE SIDE EFFECTS, REQUIRE GLOBAL
  // VDR OBJECTS TO EXIST OR PRODUCE ANY OUTPUT!
  vip1710device = NULL;
}

cPluginVip1710::~cPluginVip1710()
{
  // Clean up after yourself!
  if (vip1710device) {
    // FIXME: 'delete vip1710device;' gives segfaults
    vip1710device = NULL;
  }
}

const char * cPluginVip1710::Version(void) {
  return VERSION;
}

const char * cPluginVip1710::Description(void) {
  return tr(DESCRIPTION);
}

const char *cPluginVip1710::CommandLineHelp(void)
{
  // Return a string that describes all known command line options.
  return NULL;
}

bool cPluginVip1710::ProcessArgs(int argc, char *argv[])
{
  // Implement command line argument processing here if applicable.
  return true;
}

bool cPluginVip1710::Initialize(void)
{
  // Initialize any background activities the plugin shall perform.
#if VDRVERSNUM < 10507
  RegisterI18n(vip1710_Phrases); 
#endif //VDRVERSNUM < 10507
  vip1710device = new cVip1710Device();
  return true;
}

bool cPluginVip1710::Start(void)
{
  // Start any background activities the plugin shall perform.
  return true;
}

void cPluginVip1710::Housekeeping(void)
{
  // Perform any cleanup or other regular tasks.
}

cMenuSetupPage * cPluginVip1710::SetupMenu(void)
{
  return new cVip1710MenuSetup();
}

bool cPluginVip1710::SetupParse(const char *Name, const char *Value)
{
  if      (!strcasecmp(Name, "LogLevel"))                Vip1710Setup.LogLevel               = atoi(Value);
  else if (!strcasecmp(Name, "UseScartWSS"))             Vip1710Setup.UseScartWSS            = atoi(Value);
  else if (!strcasecmp(Name, "AC3Gain"))                 Vip1710Setup.AC3Gain                = atoi(Value);
  else if (!strcasecmp(Name, "BlackVideoForAudioOnly"))  Vip1710Setup.BlackVideoForAudioOnly = atoi(Value);
  else if (!strcasecmp(Name, "RecodeMP2AudioOnly"))      Vip1710Setup.RecodeMP2AudioOnly     = atoi(Value);
  else if (!strcasecmp(Name, "RecodeJointStereo"))       Vip1710Setup.RecodeJointStereo      = atoi(Value);
  else
    return false;
  return true;
}

const char **cPluginVip1710::SVDRPHelpPages(void)
{
  // Return help text for SVDRP commands this plugin implements
  static const char *HelpPages[] = {
    "RESET\n"
    "    Reset vip1710-plugin\n",
    NULL
    };
  return HelpPages;
}

cString cPluginVip1710::SVDRPCommand(const char *Command, const char *Option, int &ReplyCode)
{
  if(!strcasecmp(Command,"RESET")) {
    new cVip1710Device();
    return "resetted vip1710-plugin";
  }
  return NULL;
}

VDRPLUGINCREATOR(cPluginVip1710); // Don't touch this!
