/*
 * vip1710device.c: 
 *
 * See the README file for copyright information and how to reach the author.
 *
 */
#include <unistd.h>
#include <features.h>		/* Uses _GNU_SOURCE to define getsubopt in stdlib.h */
#include <inttypes.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <linux/types.h>
#include <sys/utsname.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/time.h>
#include <unistd.h>
#include <math.h>
#include <time.h>
#include <errno.h>
#include <sys/ioctl.h>

#include "vip1710device.h"
#include "vip1710blackvideo.h"

#include <vdr/dvbspu.h>
#include <vdr/status.h>
#include <vdr/device.h>
#include <vdr/tools.h>

#include "vip1710tools.h"
#include "vip1710audio.h"
#include "vip1710player.h"
#include "vip1710setup.h"

static cVip1710Device  *Instance = 0;

cVip1710Device::cVip1710Device(void)
    : fd_out(-1)
{
	int length;
	int err;

	Instance = this;

	fd_server = socket(PF_UNIX, SOCK_STREAM, 0);
	if( fd_server == -1 ) {
		log(vipERROR, "vip1710: Can't create output server socket");
		perror("vip1710-plugin: socket");
		return;
	}
	unlink(VIP1710_TS_SOCKET);
	fd_addr.sun_family = AF_UNIX;
	length = sizeof(fd_addr.sun_family) +
		sprintf(fd_addr.sun_path, VIP1710_TS_SOCKET);
	if( bind(fd_server, (struct sockaddr *) &fd_addr, length) != 0 ) {
		log(vipERROR, "vip1710: Can't bind to output server socket");
		perror("vip1710-plugin: socket");
		return;
	}
	chmod(VIP1710_TS_SOCKET, 0777);
	if( listen(fd_server, 1) != 0 ) {
		log(vipERROR, "vip1710: Can't listen to output server socket");
		perror("vip1710-plugin: socket");
		return;
	}
	fcntl(fd_server, F_SETFL, fcntl(fd_server, F_GETFL) | O_NONBLOCK);
	length = 8192;
	err = setsockopt(fd_server, SOL_SOCKET, SO_SNDBUF, (char *)&length, (int)sizeof(length));
	if(err) {
		log(vipERROR, "vip1710: Can't set output server socket options");
		perror("vip1710-plugin: setsockopt");
		return;
	}

	OpenFramebuffer();
	m_Player = new cVip1710Player();

	m_AC3toMP2Init = false;
	m_AC3DecodeStatePtr = (AC3DecodeState_t *)&m_AC3DecodeState;
	audiomode = 0; //assume stereo setting
                       //Because ivtv sets AUDIO_MONO_LEFT for bilingual streams, we rely on vdr to set SetAudioChannel() to stereo 
	framecount = 0;
	newStream = true;
	streamtype = undef;
	m_MP2RecodeInit = false;
	DecEncMP2Audio = false;

	aspectratio = 0;
}

void cVip1710Device::OpenFramebuffer()
{
        DFBSurfaceDescription dsc;
	int argc_dummy = 1;
	char *argv_dummy[] = { "/usr/bin/vdr" };
	char **argv_ptr = argv_dummy;
	osdbufsize = 720 * 576 * 4;
	osdbuffer = new unsigned char[osdbufsize];

        DFBCHECK (DirectFBInit (&argc_dummy, &argv_ptr));
        DFBCHECK (DirectFBCreate (&dfb));
        DFBCHECK (dfb->SetCooperativeLevel (dfb, DFSCL_FULLSCREEN));  

        dsc.flags = DSDESC_CAPS;
        dsc.caps  = DSCAPS_PRIMARY /* | DSCAPS_FLIPPING */;

        DFBCHECK (dfb->CreateSurface(dfb, &dsc, &dfb_primary));
        DFBCHECK (dfb_primary->Clear(dfb_primary, 0, 0, 0, 0));

	spuDecoder = NULL;
	ResetVideoSize();
}

int cVip1710Device::WriteMPEGStream(const uchar *Data, int Length, int TimeoutMs, int RetryMs)
{
	int retval;
	//log(vipDEBUG1, "got %d bytes of output, timeoutms %d, retryms %d, retval is %d", Length, TimeoutMs, RetryMs, retval);

	if(fd_out >= 0) {
		retval = WriteAllOrNothing(fd_out, Data, Length, TimeoutMs, RetryMs);
		return retval;
	} else {
		log(vipERROR, "no client to send data to!");
		errno = EAGAIN;
		return -1;
	}
}

cVip1710Device::~cVip1710Device()
{
	log(vipDEBUG1, "vip1710: ~cVip1710Device()");
	delete spuDecoder;
	if (m_AC3toMP2Init) {
		delete m_AC3toMP2;
	}
	if (m_MP2RecodeInit) {
		delete m_MP2toMP2;
	}
	if (osdbuffer != NULL) {
		log(vipDEBUG1, "vip1710: ~cVip1710Device() - delete osdbuffer");
		delete[] osdbuffer;
	}

        dfb_primary->Release( dfb_primary );
        dfb->Release( dfb );

	delete m_Player;

	if(fd_out >= 0)
		close(fd_out);
	if(fd_server >= 0)
		close(fd_server);
}

void cVip1710Device::MakePrimaryDevice(bool On)
{
	log(vipINFO, "cVip1710Device::MakePrimaryDevice(%s)", On ? "true" : "false");
	if (On) {
	  if (HasDecoder()) {
		  new cVip1710OsdProvider(dfb_primary, osdbuffer);
	  }
	}
#if VDRVERSNUM >= 10711
	cDevice::MakePrimaryDevice(On);
#endif
}

bool cVip1710Device::HasDecoder(void) const
{
	return true;

//	return fd_out >= 0; // We can decode MPEG2
}

bool cVip1710Device::CanReplay(void) const 
{
	return true;  // We can replay
}

bool cVip1710Device::SetPlayMode(ePlayMode PlayMode)
{
	log(vipDEBUG1, "cVip1710Device::SetPlayMode() (%s)", \
		PlayMode == pmNone?"pmNone":
		PlayMode == pmAudioVideo?"pmAudioVideo":
		PlayMode == pmAudioOnly?"pmAudioOnly":
		PlayMode == pmAudioOnlyBlack?"pmAudioOnlyBlack":
		PlayMode == pmVideoOnly?"pmVideoOnly":"pmExtern_THIS_SHOULD_BE_AVOIDED");
	switch (PlayMode) {
		case pmNone:		// audio/video from decoder
			streamtype=undef;
			framecount = 0;
			newStream = true;
			DecoderStop(1);
			break;
		case  pmExtern_THIS_SHOULD_BE_AVOIDED:
			close(fd_out);
			fd_out = -1;
			break;
		case pmAudioOnlyBlack:	// audio only from player, no video (black screen)
		case pmAudioOnly: 	// audio only from player, video from decoder
		case pmAudioVideo:	// audio/video from player
		default:
			DecoderPlay(0);
	}
	m_PlayMode = PlayMode;
	return true;
}

/*
	speed == 0 || speed == 1000: normal speed
	speed == 1: single step forward
	speed == -1: single step backward
	1 < speed < 1000: slow forward
	speed > 1000: fast forward
	speed == -1000: reverse play at normal speed
	-1000 < speed < -1: slow reverse
	speed < -1000: fast reverse.
 	
	other description:

	0 or 1000 specifies normal speed,
	1 specifies forward single stepping,
	-1 specifies backward single stepping,
	>1: playback at speed/1000 of the normal speed,
	<-1: reverse playback at (-speed/1000) of the normal speed. 
*/

void cVip1710Device::TrickSpeed(int Speed)
{
	// TODO: Implement this for the VIP1710

	log(vipDEBUG1, "cVip1710Device::TrickSpeed() - Set speed %d  NOT IMPLEMENTED!", Speed);

	switch (Speed) {
		case 1 : // FastForward [3>>] or FastBackward [<<3]
			DecoderPlay(1000);
			break; 
		case 3 : // FastForward [2>>] or FastBackward [<<2]
			DecoderPlay(500);
			break;
		case 6 : // FastForward [1>>] or FastBackward [<<1]
			DecoderPlay(250);
			break;
		case 8 : // SlowForward [1|>]
			DecoderPlay(40);
			break; 
		case 4 : // SlowForward [2|>]
			DecoderPlay(80);
			break; 
		case 2 : // SlowForward [3|>]
			DecoderPlay(160);
			break;
		case 63 : // SlowReverse [<|1]
			DecoderPlay(40);
			break; 
		case 48 : // SlowReverse [<|2]
			DecoderPlay(80);
			break; 
		case 24 : // SlowReverse [<|3]
			DecoderPlay(160);
			break;
	}
}

void cVip1710Device::SetVideoFormat(bool VideoFormat16_9)
{
	log(vipINFO, "cVip1710Device::SetVideoFormat(VideoFormat16_9=%s)",
		VideoFormat16_9 ? "true" : "false");
	Format16_9 = VideoFormat16_9;
}

#if APIVERSNUM >= 10708
void cVip1710Device::GetVideoSize(int &Width, int &Height, double &VideoAspect)
{
	if (fd_out >= 0) {
		Width  = horizontal_size;
		Height = vertical_size;
		switch (aspectratio) {
			default:
			case 2:   VideoAspect =  4.0 / 3.0; break;
			case 3:   VideoAspect = 16.0 / 9.0; break;
			case 4:   VideoAspect =       2.21; break;
		}
		return;
	}

	cDevice::GetVideoSize(Width, Height, VideoAspect);
	log(vipDEBUG1, "cVip1710Device::GetVideoSize: Width=%d, Height=%d, VideoAspect=%d",
		Width, Height, VideoAspect);
}

void cVip1710Device::GetOsdSize(int &Width, int &Height, double &PixelAspect)
{
	Width = 720;
	Height = 576;
  	PixelAspect = 1.0; //use default vdr value for pmExtern
	log(vipDEBUG1, "cVip1710Device::GetOsdSize: Width=%d, Height=%d, PixelAspect=%d",
		Width, Height, PixelAspect);
}
#endif

void cVip1710Device::DecoderStop(int blank)
{
	// errrmmm... for the VIP1710?
	log(vipDEBUG2, "cVip1710Device::DecoderStop(%d)", blank);
	m_Player->VidStop();
	if(fd_out >= 0) {
		close(fd_out);
		fd_out = -1;
	}
}

void cVip1710Device::DecoderPlay(int speed)
{
	int tries;
	// errrmmm... for the VIP1710?
	// maybe spawn a "toish uri LoadUri ..." here?
	log(vipDEBUG2, "cVip1710Device::DecoderPlay(%d)", speed);
	m_Player->VidPlay();
	if(fd_out < 0) {
		tries = 20; /* 20 x 50ms = 1s */
		// wait for connection
		while(tries--) {
			fd_out = accept(fd_server, 0, 0);
			if(fd_out == -1 && (errno == EAGAIN || errno == EWOULDBLOCK)) {
				usleep(50000); /* wait 50ms before retrying */
			} else {
				break;
			}
		}
	}
	if(fd_out) {
		/*fcntl(fd_out, F_SETFL, fcntl(fd_out, F_GETFL) | O_NONBLOCK);*/
		log(vipDEBUG1, "vip1710: fd_out is now %d", fd_out);
	}
}

void cVip1710Device::DecoderPaused(int paused)
{
	// errrmmm... for the VIP1710?
	log(vipDEBUG2, "cVip1710Device::DecoderPause(%d)", paused);
	m_Player->VidPause();
}

void cVip1710Device::Clear(void)
{
	log(vipDEBUG1, "cVip1710Device::Clear()");
	DecoderStop(0); //stop, clear buffers and show last frame
	DecoderPlay(0); //resume play/decoding
	cDevice::Clear();
	///< A derived class must call the base class function to make sure
	///< all registered cAudio objects are notified.
}

void cVip1710Device::Play(void)
{
	log(vipDEBUG1, "cVip1710Device::(Resume) Playback");
	DecoderPlay(1000); //normal speed
	//we cannot use VIDEO_CMD_CONTINUE: Leaving slow speed (trickmode) must reset to normal speed
}

void cVip1710Device::Freeze(void)
{
	log(vipDEBUG1, "cVip1710Device::Freeze()");
	DecoderPaused(1);
}

void cVip1710Device::Mute(void)
{
	log(vipDEBUG1, "cVip1710Device::Mute()");
	//PLAYERMUTE

	///< A derived class must call the base class function to make sure
	///< all registered cAudio objects are notified.
	cDevice::Mute();
}

void cVip1710Device::SetAudioChannelDevice(int AudioChannel)
{
	log(vipDEBUG1, "cVip1710Device::SetAudioChannelDevice (%s)  NOT IMPLEMENTED",
		AudioChannel == 0?"stereo":
		AudioChannel == 1?"mono left":"mono right");
	// 0=stereo, 1=left, 2=right, -1=no information available.
	
	audiomode = AudioChannel;
}

int cVip1710Device::GetAudioChannelDevice(void)
{
	//ivtv driver does not support AUDIO_GET_STATUS ioctl, so we return the mode which was selected at last
	log(vipDEBUG1, "cVip1710Device::GetAudioChannelDevice(): currently selected AudioChannel is %s",
		audiomode == 0?"stereo":
		audiomode == 1?"mono left":"mono right");
	return audiomode;
}

int cVip1710Device::PlayAudio(const uchar *Data, int Length, uchar Id)
{
	int PayloadOffset;
	int PES_Packet_Length;
	int len;
	static uchar Last_Id = 0x00;

	if (streamtype == undef) {
		streamtype = audio_only;
		Set_wss_mode(8); // 4:3 aspect ratio for radio
		log(vipDEBUG1, "cVip1710Device::detected Audio-only stream");
	}

#ifdef DEBUG
	#if VDRVERSNUM > 10700
        if (Vip1710Setup.LogLevel > 2) {
		PesDump(__FUNCTION__, Data, Length);
	}
	#endif
#endif

	/* look if at the beginning there is the pes packet start indicator */
	if (Data[PESPACKETSTARTCODE1] != 0x00 ||
	    Data[PESPACKETSTARTCODE2] != 0x00 ||
	    Data[PESPACKETSTARTCODE3] != 0x01) {
		log(vipDEBUG2, "cVip1710Device::PlayAudio(): no start indicator found...");
		return Length; // Return success to VDR
	}

	if (Id == 0x00) {
		/* needed for vdr >= 1.7.1 and vdr 1.6.0-2 with TSplay patch */
		Id = Data[PESSTREAMID];
	}

	/*
	 * PES Packet Length:	Specifies the number of bytes remaining
	 * 			in the packet after this field.
	 */
	len = Data[PESPACKETLEN1];
	PES_Packet_Length = (len << 8) + Data[PESPACKETLEN2];
	PES_Packet_Length += PESFIXHEADERLEN;
	if (Length != PES_Packet_Length) {
		log(vipERROR, "cVip1710Device::PlayAudio(): PES Packet seems to be corrupted");
		log(vipDEBUG1, "cVip1710Device::PlayAudio():Length=%d, len=%d, Id=0x%02x",
			Length, PES_Packet_Length, Id);
	}
#ifdef DEBUG
	MPEG2PES_t *mpeg2pes = (MPEG2PES_t *)Data;
	log(vipDEBUG2, "vip1710: PES header(Audio)" \
		" prefix=0x%02x%02x%02x" \
		" stream_id=0x%02x" \
		" packet_length=%d (Length=%d)" \
		" marker_bits=%d",
			mpeg2pes->sync1, mpeg2pes->sync2, mpeg2pes->sync3,
			mpeg2pes->stream_id,
			((mpeg2pes->packet_length_H << 8) + mpeg2pes->packet_length_L),
			Length,
			mpeg2pes->marker_bits);
	log(vipDEBUG2, "vip1710:        scrambling_control=%d" \
		" priority=%d" \
		" data_alignment_indicators=%d" \
		" copyright=%d" \
		" original=%d",
			mpeg2pes->scrambling_control,
			mpeg2pes->priority,
			mpeg2pes->data_alignment_indicators,
			mpeg2pes->copyright,
			mpeg2pes->original);
	log(vipDEBUG2, "vip1710:        PTS_DTS_indicator=%d" \
		" ESCR_flag=%d" \
		" ES_rate_flag=%d" \
		" DSM_trick_mode_flag=%d" \
		" additional_copy_info_flag=%d",
			mpeg2pes->PTS_DTS_indicator,
			mpeg2pes->ESCR_flag,
			mpeg2pes->ES_rate_flag,
			mpeg2pes->DSM_trick_mode_flag,
			mpeg2pes->additional_copy_info_flag);
	log(vipDEBUG2, "vip1710:        CRC_flag=%d" \
		" extension_flag=%d" \
		" header_length=%d",
			mpeg2pes->CRC_flag,
			mpeg2pes->extension_flag,
			mpeg2pes->header_length);
#endif

	/* #AS# ToDo:
	   Next line is for MPEG2 audio (PESFIXHEADERLEN + PESMPEG2EXTLEN) only.
	   More work is needed for MPEG1 audio.
	*/
	PayloadOffset = Data[PESHEADERLENGTH] + PESMPEG2HEADLEN;

#ifdef DEBUG
	if (Length <= PayloadOffset) {
		log(vipDEBUG1, "cVip1710Device::PlayAudio():Length=%d <= PayloadOffset=%d",
			Length, PayloadOffset);
		return Length;
	}
#endif

	if (Id != Last_Id)  {
		Last_Id = Id;
		log(vipDEBUG1, "cVip1710Device::PlayAudio(): Length=%d, Id=0x%02x",
			Length, Id);
	}

	switch (Id) {
	case 0x80 ... 0x87 :	// AC3 with vdr 1.4/1.6
	case 0xBD:		// since vdr 1.7.1 we only get the 'Privat Stream 1' Id for AC3.
		if (!m_AC3toMP2Init) {
			// Create a new decode/encode audio thread
			m_AC3toMP2 = new cAC3toMP2(m_AC3DecodeStatePtr, Vip1710Setup.AC3Gain);
			m_AC3toMP2Init = true;
		}
		ProcessAC3Audio((uint8_t *)Data, PayloadOffset, Length);
		len = Length;	// Return success to VDR
		break;
	case 0xA0 ... 0xA7 :		// LPCM
		if (!m_AC3toMP2Init) {
			// Create a new decode/encode audio thread
			m_AC3toMP2 = new cAC3toMP2(m_AC3DecodeStatePtr, Vip1710Setup.AC3Gain);
			m_AC3toMP2Init = true;
		}
		// ToDo: PCM to MP2 conversion
		len = Length;	// Return success to VDR
		break;
	case 0xC0 ... 0xDF :		// MP2
		if (m_AC3toMP2Init) {
			// Delete decode/encode audio thread
			delete m_AC3toMP2;
			m_AC3toMP2Init = false;
		}
		if (newStream) {
			newStream = false;
			//check for Joint Stereo
			len = Length - PayloadOffset;
			if (CheckMPEGAudio4JointStereo(&Data[PayloadOffset], len)) {
			    if (Vip1710Setup.RecodeJointStereo) {
				DecEncMP2Audio = true;
				log(vipDEBUG1, "cVip1710Device::PlayAudio(): recode Joint Stereo");
			    } else {
				DecEncMP2Audio = false;
				log(vipDEBUG1, "cVip1710Device::PlayAudio(): joint-stereo may sound tinny without recoding");
			    }
			} else {
			    DecEncMP2Audio = false;
			}
			//check if recoding for MP2 audio-only is selected
			if (streamtype == audio_only ||
			    m_PlayMode == pmAudioOnlyBlack ||
			    m_PlayMode == pmAudioOnly) {
				if (Vip1710Setup.RecodeMP2AudioOnly) {
					log(vipDEBUG1, "cVip1710Device::PlayAudio(): will recode MP2 audio-only");
					DecEncMP2Audio = true;
				}
				else {
					log(vipDEBUG1, "cVip1710Device::PlayAudio(): some DVB radio streams may play not properly without recoding");
					DecEncMP2Audio = false;
				}
			}
		}
		if (DecEncMP2Audio) {
			if (!m_MP2RecodeInit) {
				// Create a new MP2 decode/encode audio thread
				m_MP2toMP2 = new cMP2toMP2();
				m_MP2RecodeInit = true;
			}
			if (!ConvertMP2Audio((uint8_t *)Data, PayloadOffset, Length)) {
				// Converting the MP2 audio failed, process the 'original' data.
				ProcessMP2Audio((uint8_t *)Data, PayloadOffset, Length);
			}
		} else {
			if (m_MP2RecodeInit) {
				// Delete MP2 decode/encode audio thread
				delete m_MP2toMP2;
				m_MP2RecodeInit = false;
			}
			ProcessMP2Audio((uint8_t *)Data, PayloadOffset, Length);
		}
		len = Length;	// Return success to VDR
		break;
	default:
		static uchar LastUnknownId = 0;
		if (Id != LastUnknownId) {
			LastUnknownId = Id;
			log(vipERROR, "vip1710: PlayAudio unknown audio format, Id=0x%02x", Id);
		}
		len = Length;	// Return success to VDR
		break;
	}
	return len;
}

int cVip1710Device::PlayVideo(const uchar *Data, int Length)
{
	int len;
#ifdef DEBUG
	MPEG2PES_t *mpeg2pes = (MPEG2PES_t *)Data;
	log(vipDEBUG2, "vip1710: PES header(Video)" \
		" prefix=0x%02x%02x%02x" \
		" stream_id=0x%02x" \
		" packet_length=%d (Length=%d)" \
		" marker_bits=%d",
			mpeg2pes->sync1, mpeg2pes->sync2, mpeg2pes->sync3,
			mpeg2pes->stream_id,
			((mpeg2pes->packet_length_H << 8) + mpeg2pes->packet_length_L),
			Length,
			mpeg2pes->marker_bits);
	log(vipDEBUG2, "vip1710:        scrambling_control=%d" \
		" priority=%d" \
		" data_alignment_indicators=%d" \
		" copyright=%d" \
		" original=%d",
			mpeg2pes->scrambling_control,
			mpeg2pes->priority,
			mpeg2pes->data_alignment_indicators,
			mpeg2pes->copyright,
			mpeg2pes->original);
	log(vipDEBUG2, "vip1710:        PTS_DTS_indicator=%d" \
		" ESCR_flag=%d" \
		" ES_rate_flag=%d" \
		" DSM_trick_mode_flag=%d" \
		" additional_copy_info_flag=%d",
			mpeg2pes->PTS_DTS_indicator,
			mpeg2pes->ESCR_flag,
			mpeg2pes->ES_rate_flag,
			mpeg2pes->DSM_trick_mode_flag,
			mpeg2pes->additional_copy_info_flag);
	log(vipDEBUG2, "vip1710:        CRC_flag=%d" \
		" extension_flag=%d" \
		" header_length=%d",
			mpeg2pes->CRC_flag,
			mpeg2pes->extension_flag,
			mpeg2pes->header_length);
#endif

	/* if the first packet was an audio packet, the stream was falsely detected as audio-only and
	   a black video frame was sent. We need to Clear the decoder buffer now to avoid
	   artefacts when the next real video frame appears  */
	if (streamtype == audio_only && Vip1710Setup.BlackVideoForAudioOnly == 1) {
		log(vipDEBUG1, "cVip1710Device::PlayVideo: video packet after blackframe (audio-only) detected- reset the decoder now");
		DecoderStop(1);
		DecoderPlay(0);
	}
	streamtype = video;

	/* look if at the beginning there is the pes packet start indicator */
	if (Data[PESPACKETSTARTCODE1] != 0x00 ||
	    Data[PESPACKETSTARTCODE2] != 0x00 ||
	    Data[PESPACKETSTARTCODE3] != 0x01) {
		log(vipDEBUG2, "cVip1710Device::PlayVideo(): no start indicator found...");
		return Length; // Return success to VDR
	}

	SetVidInfo(Data, Length);
	len = WriteMPEGStream(Data, Length, 1000, 10);

	if (len <= 0 && errno != EAGAIN) { //don't log if "Resource temporarily unavailable"
		log(vipERROR, "vip1710: PlayVideo written=%d error=%d:%s",
			len, errno, strerror(errno));
	}
	return len;
}

void cVip1710Device::SetVolumeDevice(int Volume)
{
	//PLAYERSETVOL
	m_Player->SetVolume(Volume * 100 / 255);
}

void cVip1710Device::StillPicture(const uchar *Data, int Length)
{
#define MIN_IFRAME 400000
	log(vipDEBUG1, "cVip1710Device::StillPicture(Length=%d)", Length);
	if (!Data || Length < TS_SIZE) {
		return;
	}

	if (Data[PESPACKETSTARTCODE1] == 0x47) {
		// TS data
		cDevice::StillPicture(Data, Length);
		return;
	}

	/* Check if Data-Buffer contains a MPEG-1 or MPEG-2 video stream */
	if (Data[PESPACKETSTARTCODE1] == 0x00 &&
	    Data[PESPACKETSTARTCODE2] == 0x00 &&
	    Data[PESPACKETSTARTCODE3] == 0x01 &&
	    (Data[PESSTREAMID] & 0xF0) == 0xE0) {
		// PES data
		uchar *buf = MALLOC(uchar, Length);
		if (!buf) {
			return;
		}
		int i = 0;
		int blen = 0;
		while (i < Length - PESHEADERLENGTH) {
			if (Data[i + PESPACKETSTARTCODE1] == 0x00 &&
			    Data[i + PESPACKETSTARTCODE2] == 0x00 &&
			    Data[i + PESPACKETSTARTCODE3] == 0x01) {
				// Calculate PES packet length
				int len = Data[i + PESPACKETLEN1] * 256 + Data[i + PESPACKETLEN2];
				if ((Data[i + PESSTREAMID] & 0xF0) == 0xE0) {
					// MPEG-1 or MPEG-2 video packet found, skip PES header
					int offs = i + 6;
					if ((Data[i + PESEXTENSION1] & 0xC0) == 0x80) {
						// MPEG-2 PES header found, skip header extension
						if (Data[i + PESHEADERLENGTH] >= Length) {
							break;
						}
						offs += 3;
						offs += Data[i + PESHEADERLENGTH];
						len -= 3;
						len -= Data[i + PESHEADERLENGTH];
						if (len < 0 || offs + len > Length) {
							break;
						}
					} else {
						// MPEG-1 PES header found, skip header extension
						while (offs < Length && len > 0 && Data[offs] == 0xFF) {
							offs++;
							len--;
						}
						// Check for optional STD buffer size
						if (offs <= Length - 2 && len >= 2 &&
						    (Data[offs] & 0xC0) == 0x40) {
							offs += 2;
							len -= 2;
						}
						// Check for optional PTS/DTS
						if (offs <= Length - 5 && len >= 5 &&
						    (Data[offs] & 0xF0) == 0x20) {
							offs += 5;
							len -= 5;
						} else if (offs <= Length - 10 && len >= 10 &&
						           (Data[offs] & 0xF0) == 0x30) {
							offs += 10;
							len -= 10;
						} else if (offs < Length && len > 0) {
							offs++;
							len--;
						}
					}
					if (blen + len > Length) {
						// invalid PES length field
						break;
					}
					memcpy(&buf[blen], &Data[offs], len);
					i = offs + len;
					blen += len;
				} else if (Data[i + PESSTREAMID] >= 0xBD &&
					   Data[i + PESSTREAMID] <= 0xDF) {
					// other PES packets
					i += len + 6;
				} else {
					i++;
				}
			} else {
				i++;
			}
		} // End while()

		for (int i = MIN_IFRAME / Length + 1; i > 0; i--) {
			WriteMPEGStream(buf, blen);
			usleep(1); // allows the buffer to be displayed in case the progress display is active
		}
		free(buf);
	} else {
		// non-PES data
		for (int i = MIN_IFRAME / Length + 1; i > 0; i--) {
			WriteMPEGStream(Data, Length, 1000, 10);
		usleep(1); // allows the buffer to be displayed in case the progress display is active
		}
	}
}

bool cVip1710Device::Poll(cPoller &Poller, int TimeoutMs)
{
	//log(vipDEBUG2, "cVip1710Device::Poll(TimeoutMs=%d)", TimeoutMs);
	if(fd_out < 0) return false;

	Poller.Add(fd_out,true);
	return Poller.Poll(TimeoutMs);
}


cSpuDecoder *cVip1710Device::GetSpuDecoder(void)
{
	log(vipDEBUG1, "cVip1710Device::GetSpuDecoder()");
	if (!spuDecoder && IsPrimaryDevice()) {
		spuDecoder = new cDvbSpuDecoder();
	}
	return spuDecoder;
}

void cVip1710Device::SetVidInfo(const uchar *mbuf, int count)
{
	// TODO: implement SCART 16:9 signalling?

	const uchar *headr;
	int found = 0;
	int c = 0;
	int newaspectratio;

	while (found < 4 && c+4 < count){
		const uchar *b;

		b = mbuf + c;
		if (b[0] == 0x00 && b[1] == 0x00 &&
		    b[2] == 0x01 && b[3] == 0xb3) {
			found = 4;
		} else {
			c++;
		}
	}

	if (! found) return;
	c += 4;
	if (c+12 >= count) return;
	headr = mbuf+c; 

	horizontal_size = ((headr[1] &0xF0) >> 4) | (headr[0] << 4);
	vertical_size   = ((headr[1] &0x0F) << 8) | (headr[2]);
	newaspectratio = (int)((headr[3]&0xF0) >> 4) ;

	if(aspectratio != newaspectratio) {
		aspectratio = newaspectratio;
		log(vipDEBUG2, "cVip1710Device::SetVidInfo() aspectratio=%s",
			aspectratio == 1?"1:1":
			aspectratio == 2?"4:3":
			aspectratio == 3?"16:9":
			aspectratio == 4?"2:21:1":"unknown");
		m_Player->SetAspectRatio(aspectratio, Vip1710Setup.UseScartWSS);
	}

#if 0
	if ((horizontal_size > 720) || (vertical_size > 576))
		return;
#endif
}

int64_t cVip1710Device::GetSTC(void)
{
	log(vipDEBUG2, "cVip1710Device::GetSTC");
	// TODO: implement something?
	return -1;
}

void cVip1710Device::SetVideoSize(int x, int y, int w, int d)
{
	log(vipDEBUG1, "cVip1710Device::SetVideoSize(x=%d,y=%d,w=%d,d=%d)", x, y, w, d);
/*
	width = w;
	height = d;
	lx = x;
	ly = y;
	sizechanged = 1;
*/
}

void cVip1710Device::Set_wss_mode(int wss_data)
{
	log(vipDEBUG1, "cVip1710Device::Set_wss_mode(wss_data=%d)", wss_data);
}

void SetVideoSize(int x, int y, int w, int d)
{
	Instance->SetVideoSize(x, y, w, d);
}

void ResetVideoSize()
{
	/* FIXME: Think NTSC */
	Instance->SetVideoSize(0, 0, 720, 576);
}

void cVip1710Device::ProcessAC3Audio(uint8_t *PESPacket, int PayloadOffset, int Length)
{
	int len;
	int bytes_left;
	int pcm_bytes;
	int mp2_len;
	int mp2_bytes_encoded;
	uint8_t *PESPacketData;
	int PESPacketData_len;
	uint8_t *MP2PESDataPtr;
	int PES_Packet_Length;
	uint8_t *DataPtr;
	int16_t *PCMBufferPtr;
	int mp2buffer_size;

	if (!m_AC3toMP2Init) {
		return;		// AC3 audio thread not running
	}

	log(vipDEBUG2, "cVip1710Device::ProcessAC3Audio()");

	PESPacketData = (uint8_t *)&PESPacket[PayloadOffset];
	PESPacketData_len = Length - PayloadOffset;
	bytes_left = PESPacketData_len;
	DataPtr = PESPacketData;
	PCMBufferPtr = (int16_t *)m_PCMBuffer;
	mp2_bytes_encoded = 0;
	mp2buffer_size = MP2_PES_DATA_SIZE;

	// copy 'original' PES header to MP2 PES buffer
	memcpy(m_MP2PESBuffer, PESPacket, PayloadOffset);
	// Initialize output pointer for encode call
	MP2PESDataPtr = (uint8_t *)&m_MP2PESBuffer[PayloadOffset];

	while (bytes_left > 0) {
		len = m_AC3toMP2->A52DecodeFrame(m_AC3DecodeStatePtr,
						PCMBufferPtr, &pcm_bytes,
						DataPtr, bytes_left);
		if (len < 0) {
			log(vipERROR, "vip1710: Error return from A52DecodeFrame()");
			return;
		}
		DataPtr += len;		// Adjust input pointer for next decode call
		bytes_left -= len;
		if (pcm_bytes) {
			mp2_len = m_AC3toMP2->MP2EncodeFrame(PCMBufferPtr,
								pcm_bytes,
								MP2PESDataPtr,
								mp2buffer_size);
			if (mp2_len < 0) {
				log(vipERROR, "vip1710: ProcessAC3Audio() - MP2PESBuffer too small");
				break;
			}
			mp2_bytes_encoded += mp2_len;
			MP2PESDataPtr += mp2_len;	// Adjust outbuf pointer for next encode call
			mp2buffer_size -= mp2_len;	// Remaining bytes in outbuf buffer
		}
	}

#if 0 // if'ed out in original pvr350 sources!
	/*
	 * Flush any remaining audio. (don't send any new audio data)
	 * There should only ever be a max of 1 frame on a flush. There may be zero frames
	 * if the audio data was an exact multiple of 1152 (TWOLAME_SAMPLES_PER_FRAME)
	*/
	mp2_len = m_MP2toMP2->MP2EncodeFrame(NULL,
						0,
						MP2PESDataPtr,
						mp2buffer_size);
	if (mp2_len > 0) {
		log(vipDEBUG1, "ProcessAC3Audio() - flush remaining data [%d]", mp2_len);
		mp2_bytes_encoded += mp2_len;
	}
#endif

	if (mp2_bytes_encoded > 0) {
		// Change PES Stream ID
		m_MP2PESBuffer[PESSTREAMID] = MPEG_AUDIO_STREAM_ID;

		// Adjust PES Packet Length
		PES_Packet_Length = PayloadOffset - PESFIXHEADERLEN + mp2_bytes_encoded;
		m_MP2PESBuffer[PESPACKETLEN1] = (uint8_t)(PES_Packet_Length >> 8);
		m_MP2PESBuffer[PESPACKETLEN2] = (uint8_t)(PES_Packet_Length & 0xff);

		// Calculate length of PES Packet Frame
		len = PES_Packet_Length + PESFIXHEADERLEN;

		len = WriteMPEGStream(m_MP2PESBuffer, len, 1000, 10);
		if (len <= 0 && errno != EAGAIN) { //don't log if "Resource temporarily unavailable"
			log(vipERROR, "vip1710: ProcessAC3Audio written=%d error=%d:%s",
				len, errno, strerror(errno));
		}
	}	
	return;
}

bool cVip1710Device::CheckMPEGAudio4JointStereo(const uchar *PESData, int Length)
{
	int len = 0;
	uint8_t *Data;
	uchar ID;
	uchar Layer;
	uchar ChannelMode;
	uchar ModeExtension;

	if (Length < SUBSTREAM_MP2_LENGTH) {
		// Not enough bytes left for MP3 Audio Substream Header
		log(vipDEBUG1, "CheckMPEGAudio4JointStereo() - Not enough bytes for MP3 Audio Header !");
		return false;
	}

	while (len < Length) {
		Data = (uint8_t *)(PESData + len);

		if ((Length - len) < SUBSTREAM_MP2_LENGTH) {
			// Not enough bytes left for MP3 Audio Substream Header
			log(vipDEBUG1, "CheckMPEGAudio4JointStereo() - No MP3 Audio Header found !");
			return false;
		}

		/*
		 * Audio Frame Header Description
		 * ==============================
		 * Syncword (11 bits all set to 1)
		 * ID (2 bits)
		 *	0 = MPEG Version 2.5
		 * 	1 = reserved
		 *	2 = MPEG Version 2
		 *	3 = MPEG Version 1
		 * Layer (2 bits)
		 *	0 = reserved
		 * 	1 = Layer III
		 *	2 = Layer II
		 *	3 = Layer I
		 * Protection (1 bit)
		 * Bitrate (4 bits; 1111 => not allowed)
		 * Frequency (2 bits; 11 => reserved)
		 * Padding (1 bit)
		 * Private (1 bit)
		 * Channel Mode (2 bits)
		 *	0 = Stereo
		 * 	1 = Joint Stereo
		 *	2 = Two Mono Channels
		 *	3 = One Mono Channel
		 * Mode Extension (2 bits)
		 * Copyright (1 bit)
		 * Original (1 bit)
		 * Emphasis (2 bits)
		 */
		if ((Data[SUBSTREAM_MP2_SYNC] == 0xff) &&
		    ((Data[SUBSTREAM_MP2_VERSION] & 0xe0) == 0xe0) &&
		    ((Data[SUBSTREAM_MP2_VERSION] & 0x18) != 0x08) &&
		    ((Data[SUBSTREAM_MP2_VERSION] & 0x06) != 0x00) &&
		    ((Data[SUBSTREAM_MP2_BITRATE] & 0xf0) != 0xf0) &&
		    ((Data[SUBSTREAM_MP2_BITRATE] & 0x0c) != 0x0c)) {
			break;
		}
		len++;
	}
	
	ID = (Data[SUBSTREAM_MP2_VERSION] & 0x18) >> 3;
	Layer = (Data[SUBSTREAM_MP2_VERSION] & 0x06) >> 1;
	ChannelMode = (Data[SUBSTREAM_MP2_MODE] & 0xc0) >> 6;
	ModeExtension = (Data[SUBSTREAM_MP2_MODE] & 0x30) >> 4;
	log(vipDEBUG2, "CheckMPEGAudio4JointStereo() ID=%d, Layer=%d, ChanMode=%d, ModeExt=%d",
			ID, Layer, ChannelMode, ModeExtension);

	// Check for channel mode for Joint Stereo
	if (ChannelMode == 0x01) {
		log(vipDEBUG1, "CheckMPEGAudio4JointStereo() - Joint Stereo detected");
		return true;
	} else {
	  	log(vipDEBUG1, "CheckMPEGAudio4JointStereo() - no Joint Stereo detected");
		return false;
	}
}

bool cVip1710Device::ConvertMP2Audio(uint8_t *Data, int PayloadOffset, int Length)
{
	int len;
	int bytes_left;
	int pcm_bytes;
	int mp2_len;
	int mp2_bytes_encoded;
	uint8_t *PESPacketData;
	int PESPacketData_len;
	uint8_t *MP2PESDataPtr;
	int PES_Packet_Length;
	uint8_t *DataPtr;
	int16_t *PCMBufferPtr;
	int mp2buffer_size;

	if (!m_MP2RecodeInit) {
		return false;	// MP2 audio thread not running
	}

	log(vipDEBUG2, "cVip1710Device::ConvertMP2Audio()");

	PESPacketData = (uint8_t *)&Data[PayloadOffset];
	PESPacketData_len = Length - PayloadOffset;

	DataPtr = PESPacketData;
	PCMBufferPtr = (int16_t *)m_PCMBuffer;
	mp2_bytes_encoded = 0;
	mp2buffer_size = MP2_PES_DATA_SIZE;

	// copy 'original' PES header to MP2 PES buffer
	memcpy(m_MP2PESBuffer, Data, PayloadOffset);
	// Initialize pointer for encode call
	MP2PESDataPtr = (uint8_t *)&m_MP2PESBuffer[PayloadOffset];

	bytes_left = m_MP2toMP2->MPG123DecodeFrame(DataPtr, PESPacketData_len,
						PCMBufferPtr, PCM_BUFFER_SIZE,
						&pcm_bytes);
	if (bytes_left < 0) {
		log(vipERROR, "vip1710: ConvertMP2Audio() - error decode audio");
		return false;
	}
	if (pcm_bytes <= 0) {
		return true;
	}
	do {
		mp2_len = m_MP2toMP2->MP2EncodeFrame(PCMBufferPtr,
							pcm_bytes,
							MP2PESDataPtr,
							mp2buffer_size);
		if (mp2_len < 0) {
			log(vipERROR, "vip1710: ConvertMP2Audio() - MP2PESBuffer too small");
			return false;
		}
		mp2buffer_size -= mp2_len;
		mp2_bytes_encoded += mp2_len;
		// Adjust pointer for next encode call
		MP2PESDataPtr += mp2_len;

		// Get all decoded audio that is available now before feeding more input.
		bytes_left = m_MP2toMP2->MPG123DecodeFrame(NULL, 0,
							PCMBufferPtr,
							PCM_BUFFER_SIZE,
							&pcm_bytes);
		if (bytes_left < 0) {
			log(vipERROR, "vip1710: ConvertMP2Audio() - error get all decoded audio");
			return false;
		}
	} while (bytes_left);

	/*
	 * Flush any remaining audio. (don't send any new audio data)
	 * There should only ever be a max of 1 frame on a flush. There may be zero frames
	 * if the audio data was an exact multiple of 1152 (TWOLAME_SAMPLES_PER_FRAME)
	*/
	mp2_len = m_MP2toMP2->MP2EncodeFrame(NULL,
						0,
						MP2PESDataPtr,
						mp2buffer_size);
	if (mp2_len > 0) {
		log(vipDEBUG1, "ConvertMP2Audio() - flush remaining data [%d]", mp2_len);
		mp2_bytes_encoded += mp2_len;
	}

	if (mp2_bytes_encoded > 0) {
		// Adjust PES Packet Length
		PES_Packet_Length = PayloadOffset - PESFIXHEADERLEN + mp2_bytes_encoded;
		m_MP2PESBuffer[PESPACKETLEN1] = (uint8_t)(PES_Packet_Length >> 8);
		m_MP2PESBuffer[PESPACKETLEN2] = (uint8_t)(PES_Packet_Length & 0xff);

		// Calculate length of PES Packet Frame
		len = PES_Packet_Length + PESFIXHEADERLEN;
		ProcessMP2Audio((uint8_t *)m_MP2PESBuffer, PayloadOffset, len);
	}

	return true;
}

void cVip1710Device::ProcessMP2Audio(uint8_t *Data, int PayloadOffset, int Length)
{
	int len;

	//log(vipDEBUG2, "cVip1710Device::ProcessMP2Audio()");

	if (streamtype == audio_only || m_PlayMode == pmAudioOnlyBlack || m_PlayMode == pmAudioOnly) {
		if (Vip1710Setup.BlackVideoForAudioOnly == 1 || m_PlayMode == pmAudioOnlyBlack) {
			//PesDump(__FUNCTION__, Data, Length);
			len = WriteMPEGStream(Data, Length, 1000, 10);
			if (len <= 0 && errno !=11) { //don't log if "Resource temporarily unavailable"
				log(vipERROR, "vip1710: ProcessMP2Audio(MP2 Audio only) written=%d error=%d:%s",
				    len, errno, strerror(errno));
			} else {
				framecount += 24; //assume frame length of 24ms for MP2 audio
			}
			if (framecount >= 40) {
				int header_size = Data[PESHEADERLENGTH] + PESMPEG2HEADLEN;
				uchar *dst = blackvideo + PESMPEG2HEADLEN + 255 - header_size;
				memcpy(dst, Data, header_size);
				dst[3] = 0xe0;
				dst[4] = 0x00;
				dst[5] = 0x00;
				int frame_length = sizeof(blackvideo) - PESMPEG2HEADLEN - 255 + header_size;
				//PesDump(__FUNCTION__, dst, frame_length);
				len = WriteMPEGStream(dst, frame_length, 1000, 10);
				if (len <= 0 && errno != EAGAIN) { //don't log if "Resource temporarily unavailable"
					log(vipERROR, "vip1710: ProcessMP2Audio(Blackframe) written=%d error=%d:%s",
						len, errno, strerror(errno));
				} else {
					framecount -= 40; //video frame length = 40 ms
				}
			}
		} else {
			len = WriteMPEGStream(Data, Length, 1000, 10);
			if (len <= 0 && errno != EAGAIN) { //don't log if "Resource temporarily unavailable"
				log(vipERROR, "vip1710: ProcessMP2Audio(MP2 Audio only) written=%d error=%d:%s",
					len, errno, strerror(errno));
			}
		}
	} else { //Audio packet for a stream that has also video
		len = WriteMPEGStream(Data, Length, 1000, 10);
		if (len <= 0 && errno != EAGAIN) { //don't log if "Resource temporarily unavailable"
			log(vipERROR, "vip1710: ProcessMP2Audio(MP2) written=%d error=%d:%s",
				len, errno, strerror(errno));
		}
	}
}

